@@ -19,9 +19,9 @@ package resources
 import (
 	"fmt"
 	"testing"
+	"time"
 
 	"github.com/google/go-cmp/cmp"
-	"github.com/google/go-cmp/cmp/cmpopts"
 
 	appsv1 "k8s.io/api/apps/v1"
 	corev1 "k8s.io/api/core/v1"
@@ -29,26 +29,36 @@ import (
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/util/intstr"
 
+	"knative.dev/networking/pkg/apis/networking"
 	"knative.dev/pkg/metrics"
 	_ "knative.dev/pkg/metrics/testing"
 	"knative.dev/pkg/ptr"
 	"knative.dev/pkg/system"
 	_ "knative.dev/pkg/system/testing"
-	"knative.dev/serving/pkg/apis/networking"
+	"knative.dev/serving/pkg/apis/autoscaling"
 	"knative.dev/serving/pkg/apis/serving"
 	v1 "knative.dev/serving/pkg/apis/serving/v1"
-	autoscalerconfig "knative.dev/serving/pkg/autoscaler/config"
+	asconfig "knative.dev/serving/pkg/autoscaler/config"
+	"knative.dev/serving/pkg/deployment"
 	"knative.dev/serving/pkg/network"
+
+	. "knative.dev/serving/pkg/testing/v1"
 )
 
 var (
-	containerName                = "my-container-name"
+	servingContainerName         = "serving-container"
+	sidecarContainerName         = "sidecar-container-1"
+	sidecarContainerName2        = "sidecar-container-2"
 	sidecarIstioInjectAnnotation = "sidecar.istio.io/inject"
-	defaultUserContainer         = &corev1.Container{
-		Name:                     containerName,
-		Image:                    "busybox",
-		Ports:                    buildContainerPorts(v1.DefaultUserPort),
-		VolumeMounts:             []corev1.VolumeMount{varLogVolumeMount},
+	defaultServingContainer      = &corev1.Container{
+		Name:  servingContainerName,
+		Image: "busybox",
+		Ports: buildContainerPorts(v1.DefaultUserPort),
+		VolumeMounts: []corev1.VolumeMount{{
+			Name:        varLogVolume.Name,
+			MountPath:   "/var/log",
+			SubPathExpr: "$(K_INTERNAL_POD_NAMESPACE)_$(K_INTERNAL_POD_NAME)_" + servingContainerName,
+		}},
 		Lifecycle:                userLifecycle,
 		TerminationMessagePolicy: corev1.TerminationMessageFallbackToLogsOnError,
 		Stdin:                    false,
@@ -63,12 +73,18 @@ var (
 			Name: "K_CONFIGURATION",
 		}, {
 			Name: "K_SERVICE",
+		}, {
+			Name:      "K_INTERNAL_POD_NAME",
+			ValueFrom: &corev1.EnvVarSource{FieldRef: &corev1.ObjectFieldSelector{FieldPath: "metadata.name"}},
+		}, {
+			Name:      "K_INTERNAL_POD_NAMESPACE",
+			ValueFrom: &corev1.EnvVarSource{FieldRef: &corev1.ObjectFieldSelector{FieldPath: "metadata.namespace"}},
 		}},
 	}
 
 	defaultQueueContainer = &corev1.Container{
 		Name:      QueueContainerName,
-		Resources: createQueueResources(make(map[string]string), &corev1.Container{}),
+		Resources: createQueueResources(&deploymentConfig, make(map[string]string), &corev1.Container{}),
 		Ports:     append(queueNonServingPorts, queueHTTPPort),
 		ReadinessProbe: &corev1.Probe{
 			Handler: corev1.Handler{
@@ -76,7 +92,7 @@ var (
 					Command: []string{"/ko-app/queue", "-probe-period", "0"},
 				},
 			},
-			PeriodSeconds:  1,
+			PeriodSeconds:  10,
 			TimeoutSeconds: 10,
 		},
 		SecurityContext: queueSecurityContext,
@@ -136,7 +152,7 @@ var (
 			Value: "false",
 		}, {
 			Name:  "TRACING_CONFIG_SAMPLE_RATE",
-			Value: "0.000000",
+			Value: "0",
 		}, {
 			Name:  "USER_PORT",
 			Value: "8080",
@@ -146,21 +162,6 @@ var (
 		}, {
 			Name:  "METRICS_DOMAIN",
 			Value: metrics.Domain(),
-		}, {
-			Name:  "USER_CONTAINER_NAME",
-			Value: containerName,
-		}, {
-			Name:  "ENABLE_VAR_LOG_COLLECTION",
-			Value: "false",
-		}, {
-			Name:  "VAR_LOG_VOLUME_NAME",
-			Value: varLogVolumeName,
-		}, {
-			Name:  "INTERNAL_VOLUME_PATH",
-			Value: internalVolumePath,
-		}, {
-			Name:  "DOWNWARD_API_LABELS_PATH",
-			Value: fmt.Sprintf("%s/%s", podInfoVolumePath, metadataLabelsPath),
 		}, {
 			Name:  "SERVING_READINESS_PROBE",
 			Value: fmt.Sprintf(`{"tcpSocket":{"port":%d,"host":"127.0.0.1"}}`, v1.DefaultUserPort),
@@ -204,7 +205,7 @@ var (
 					serving.RevisionUID: "1234",
 				},
 			},
-			ProgressDeadlineSeconds: ptr.Int32(ProgressDeadlineSeconds),
+			ProgressDeadlineSeconds: ptr.Int32(0),
 			Template: corev1.PodTemplateSpec{
 				ObjectMeta: metav1.ObjectMeta{
 					Labels: map[string]string{
@@ -218,22 +219,48 @@ var (
 			},
 		},
 	}
+)
 
-	defaultRevision = &v1.Revision{
+func defaultSidecarContainer(containerName string) *corev1.Container {
+	return &corev1.Container{
+		Name:  containerName,
+		Image: "ubuntu",
+		VolumeMounts: []corev1.VolumeMount{{
+			Name:        varLogVolume.Name,
+			MountPath:   "/var/log",
+			SubPathExpr: "$(K_INTERNAL_POD_NAMESPACE)_$(K_INTERNAL_POD_NAME)_" + containerName,
+		}},
+		Lifecycle:                userLifecycle,
+		TerminationMessagePolicy: corev1.TerminationMessageFallbackToLogsOnError,
+		Stdin:                    false,
+		TTY:                      false,
+		Env: []corev1.EnvVar{{
+			Name:  "K_REVISION",
+			Value: "bar",
+		}, {
+			Name: "K_CONFIGURATION",
+		}, {
+			Name: "K_SERVICE",
+		}, {
+			Name:      "K_INTERNAL_POD_NAME",
+			ValueFrom: &corev1.EnvVarSource{FieldRef: &corev1.ObjectFieldSelector{FieldPath: "metadata.name"}},
+		}, {
+			Name:      "K_INTERNAL_POD_NAMESPACE",
+			ValueFrom: &corev1.EnvVarSource{FieldRef: &corev1.ObjectFieldSelector{FieldPath: "metadata.namespace"}},
+		}},
+	}
+}
+
+func defaultRevision() *v1.Revision {
+	return &v1.Revision{
 		ObjectMeta: metav1.ObjectMeta{
 			UID: "1234",
 		},
 		Spec: v1.RevisionSpec{
 			TimeoutSeconds: ptr.Int64(45),
-			PodSpec: corev1.PodSpec{
-				Containers: []corev1.Container{{
-					Name:  containerName,
-					Image: "busybox",
-				}},
-			},
 		},
 	}
-)
+}
 
 func refInt64(num int64) *int64 {
 	return &num
@@ -242,7 +269,6 @@ func refInt64(num int64) *int64 {
 type containerOption func(*corev1.Container)
 type podSpecOption func(*corev1.PodSpec)
 type deploymentOption func(*appsv1.Deployment)
-type revisionOption func(*v1.Revision)
 
 func container(container *corev1.Container, opts ...containerOption) corev1.Container {
 	for _, option := range opts {
@@ -251,8 +277,12 @@ func container(container *corev1.Container, opts ...containerOption) corev1.Cont
 	return *container
 }
 
-func userContainer(opts ...containerOption) corev1.Container {
-	return container(defaultUserContainer.DeepCopy(), opts...)
+func servingContainer(opts ...containerOption) corev1.Container {
+	return container(defaultServingContainer.DeepCopy(), opts...)
+}
+
+func sidecarContainer(containerName string, opts ...containerOption) corev1.Container {
+	return container(defaultSidecarContainer(containerName).DeepCopy(), opts...)
 }
 
 func queueContainer(opts ...containerOption) corev1.Container {
@@ -275,48 +305,32 @@ func withEnvVar(name, value string) containerOption {
 	}
 }
 
-func withInternalVolumeMount() containerOption {
-	return func(container *corev1.Container) {
-		container.VolumeMounts = append(container.VolumeMounts, internalVolumeMount)
-	}
+func withTCPReadinessProbe(port int) *corev1.Probe {
+	return &corev1.Probe{
+		Handler: corev1.Handler{
+			TCPSocket: &corev1.TCPSocketAction{
+				Host: "127.0.0.1",
+				Port: intstr.FromInt(port),
+			}}}
 }
 
-func withPodLabelsVolumeMount() containerOption {
-	return func(container *corev1.Container) {
-		container.VolumeMounts = append(container.VolumeMounts, labelVolumeMount)
-	}
+func withHTTPReadinessProbe(port int) *corev1.Probe {
+	return &corev1.Probe{
+		Handler: corev1.Handler{
+			HTTPGet: &corev1.HTTPGetAction{
+				Port: intstr.FromInt(port),
+				Path: "/",
+			},
+		}}
 }
 
-func withReadinessProbe(handler corev1.Handler) containerOption {
-	return func(container *corev1.Container) {
-		container.ReadinessProbe = &corev1.Probe{Handler: handler}
-	}
-}
-
-func withTCPReadinessProbe() containerOption {
-	return withReadinessProbe(corev1.Handler{
-		TCPSocket: &corev1.TCPSocketAction{
-			Host: "127.0.0.1",
-			Port: intstr.FromInt(v1.DefaultUserPort),
-		},
-	})
-}
-
-func withHTTPReadinessProbe(port int) containerOption {
-	return withReadinessProbe(corev1.Handler{
-		HTTPGet: &corev1.HTTPGetAction{
-			Port: intstr.FromInt(port),
-			Path: "/",
-		},
-	})
-}
-
-func withExecReadinessProbe(command []string) containerOption {
-	return withReadinessProbe(corev1.Handler{
-		Exec: &corev1.ExecAction{
-			Command: command,
-		},
-	})
+func withExecReadinessProbe(command []string) *corev1.Probe {
+	return &corev1.Probe{
+		Handler: corev1.Handler{
+			Exec: &corev1.ExecAction{
+				Command: command,
+			},
+		}}
 }
 
 func withLivenessProbe(handler corev1.Handler) containerOption {
@@ -348,7 +362,7 @@ func withAppendedVolumes(volumes ...corev1.Volume) podSpecOption {
 	}
 }
 
-func makeDeployment(opts ...deploymentOption) *appsv1.Deployment {
+func appsv1deployment(opts ...deploymentOption) *appsv1.Deployment {
 	deploy := defaultDeployment.DeepCopy()
 	for _, option := range opts {
 		option(deploy)
@@ -356,8 +370,8 @@ func makeDeployment(opts ...deploymentOption) *appsv1.Deployment {
 	return deploy
 }
 
-func revision(name, ns string, opts ...revisionOption) *v1.Revision {
-	revision := defaultRevision.DeepCopy()
+func revision(name, ns string, opts ...RevisionOption) *v1.Revision {
+	revision := defaultRevision()
 	revision.ObjectMeta.Name = name
 	revision.ObjectMeta.Namespace = ns
 	for _, option := range opts {
@@ -366,7 +380,7 @@ func revision(name, ns string, opts ...revisionOption) *v1.Revision {
 	return revision
 }
 
-func withContainerConcurrency(cc int64) revisionOption {
+func withContainerConcurrency(cc int64) RevisionOption {
 	return func(revision *v1.Revision) {
 		revision.Spec.ContainerConcurrency = &cc
 	}
@@ -376,7 +390,7 @@ func withoutLabels(revision *v1.Revision) {
 	revision.ObjectMeta.Labels = map[string]string{}
 }
 
-func withOwnerReference(name string) revisionOption {
+func withOwnerReference(name string) RevisionOption {
 	return func(revision *v1.Revision) {
 		revision.ObjectMeta.OwnerReferences = []metav1.OwnerReference{{
 			APIVersion:         v1.SchemeGroupVersion.String(),
@@ -388,31 +402,96 @@ func withOwnerReference(name string) revisionOption {
 	}
 }
 
+// withContainers function helps to provide a podSpec to the revision for both single and multiple containers
+func withContainers(containers []corev1.Container) RevisionOption {
+	return func(revision *v1.Revision) {
+		revision.Spec = v1.RevisionSpec{
+			PodSpec: corev1.PodSpec{
+				Containers: containers,
+			},
+			TimeoutSeconds: ptr.Int64(45),
+		}
+	}
+}
+
+func getContainer() containerOption {
+	return func(container *corev1.Container) {
+		container.Command = []string{"/bin/bash"}
+		container.Args = []string{"-c", "echo Hello world"}
+		container.Env = append([]corev1.EnvVar{{
+			Name:  "FOO",
+			Value: "bar",
+		}, {
+			Name:  "BAZ",
+			Value: "blah",
+		}}, container.Env...)
+		container.Resources = corev1.ResourceRequirements{
+			Requests: corev1.ResourceList{
+				corev1.ResourceMemory: resource.MustParse("666Mi"),
+				corev1.ResourceCPU:    resource.MustParse("666m"),
+			},
+			Limits: corev1.ResourceList{
+				corev1.ResourceMemory: resource.MustParse("888Mi"),
+				corev1.ResourceCPU:    resource.MustParse("888m"),
+			},
+		}
+		container.TerminationMessagePolicy = corev1.TerminationMessageReadFile
+	}
+}
+
+func updateContainer() containerOption {
+	return func(container *corev1.Container) {
+		container.Command = []string{"/bin/bash"}
+		container.Args = []string{"-c", "echo Hello world"}
+		container.Env = append([]corev1.EnvVar{{
+			Name:  "FOO",
+			Value: "bar",
+		}, {
+			Name:  "BAZ",
+			Value: "blah",
+		}}, container.Env...)
+		container.Resources = corev1.ResourceRequirements{
+			Requests: corev1.ResourceList{
+				corev1.ResourceMemory: resource.MustParse("666Mi"),
+				corev1.ResourceCPU:    resource.MustParse("666m"),
+			},
+			Limits: corev1.ResourceList{
+				corev1.ResourceMemory: resource.MustParse("888Mi"),
+				corev1.ResourceCPU:    resource.MustParse("888m"),
+			},
+		}
+		container.TerminationMessagePolicy = corev1.TerminationMessageReadFile
+	}
+}
+
 func TestMakePodSpec(t *testing.T) {
 	tests := []struct {
 		name string
 		rev  *v1.Revision
 		oc   metrics.ObservabilityConfig
-		ac   autoscalerconfig.Config
 		want *corev1.PodSpec
 	}{{
 		name: "user-defined user port, queue proxy have PORT env",
 		rev: revision("bar", "foo",
-			withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				revision.Spec.GetContainer().Ports = []corev1.ContainerPort{{
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
 					ContainerPort: 8888,
-				}}
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
-			},
+				}},
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
+			withContainerConcurrency(1),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
+				servingContainer(
 					func(container *corev1.Container) {
 						container.Ports[0].ContainerPort = 8888
+						container.Image = "busybox@sha256:deadbeef"
 					},
 					withEnvVar("PORT", "8888"),
 				),
@@ -420,23 +499,27 @@ func TestMakePodSpec(t *testing.T) {
 					withEnvVar("CONTAINER_CONCURRENCY", "1"),
 					withEnvVar("USER_PORT", "8888"),
 					withEnvVar("SERVING_READINESS_PROBE", `{"tcpSocket":{"port":8888,"host":"127.0.0.1"}}`),
-				),
-			}),
+				)}),
 	}, {
 		name: "volumes passed through",
 		rev: revision("bar", "foo",
-			withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				revision.Spec.GetContainer().Ports = []corev1.ContainerPort{{
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
 					ContainerPort: 8888,
-				}}
-				revision.Spec.GetContainer().VolumeMounts = []corev1.VolumeMount{{
+				}},
+				VolumeMounts: []corev1.VolumeMount{{
 					Name:      "asdf",
 					MountPath: "/asdf",
-				}}
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
+				}},
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
+			withContainerConcurrency(1),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+			func(revision *v1.Revision) {
 				revision.Spec.Volumes = []corev1.Volume{{
 					Name: "asdf",
 					VolumeSource: corev1.VolumeSource{
@@ -449,9 +532,10 @@ func TestMakePodSpec(t *testing.T) {
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
+				servingContainer(
 					func(container *corev1.Container) {
 						container.Ports[0].ContainerPort = 8888
+						container.Image = "busybox@sha256:deadbeef"
 					},
 					withEnvVar("PORT", "8888"),
 					withPrependedVolumeMounts(corev1.VolumeMount{
@@ -475,16 +559,21 @@ func TestMakePodSpec(t *testing.T) {
 	}, {
 		name: "concurrency=1 no owner",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
 			withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
-			},
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(),
+				servingContainer(func(container *corev1.Container) {
+					container.Image = "busybox@sha256:deadbeef"
+				}),
 				queueContainer(
 					withEnvVar("CONTAINER_CONCURRENCY", "1"),
 				),
@@ -492,19 +581,19 @@ func TestMakePodSpec(t *testing.T) {
 	}, {
 		name: "concurrency=1 no owner digest resolved",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
 			withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				revision.Status = v1.RevisionStatus{
-					ImageDigest: "busybox@sha256:deadbeef",
-				}
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
-			},
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(func(container *corev1.Container) {
+				servingContainer(func(container *corev1.Container) {
 					container.Image = "busybox@sha256:deadbeef"
 				}),
 				queueContainer(
@@ -514,17 +603,22 @@ func TestMakePodSpec(t *testing.T) {
 	}, {
 		name: "concurrency=1 with owner",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
 			withContainerConcurrency(1),
 			withOwnerReference("parent-config"),
-			func(revision *v1.Revision) {
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
-			},
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(),
+				servingContainer(func(container *corev1.Container) {
+					container.Image = "busybox@sha256:deadbeef"
+				}),
 				queueContainer(
 					withEnvVar("SERVING_CONFIGURATION", "parent-config"),
 					withEnvVar("CONTAINER_CONCURRENCY", "1"),
@@ -532,14 +626,21 @@ func TestMakePodSpec(t *testing.T) {
 			}),
 	}, {
 		name: "with http readiness probe",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withHTTPReadinessProbe(v1.DefaultUserPort),
-			)
-		}),
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withHTTPReadinessProbe(v1.DefaultUserPort),
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(),
+				servingContainer(func(container *corev1.Container) {
+					container.Image = "busybox@sha256:deadbeef"
+				}),
 				queueContainer(
 					withEnvVar("CONTAINER_CONCURRENCY", "0"),
 					withEnvVar("SERVING_READINESS_PROBE", `{"httpGet":{"path":"/","port":8080,"host":"127.0.0.1","scheme":"HTTP","httpHeaders":[{"name":"K-Kubelet-Probe","value":"queue"}]}}`),
@@ -547,19 +648,21 @@ func TestMakePodSpec(t *testing.T) {
 			}),
 	}, {
 		name: "with tcp readiness probe",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withReadinessProbe(corev1.Handler{
-					TCPSocket: &corev1.TCPSocketAction{
-						Host: "127.0.0.1",
-						Port: intstr.FromInt(12345),
-					},
-				}),
-			)
-		}),
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(),
+				servingContainer(func(container *corev1.Container) {
+					container.Image = "busybox@sha256:deadbeef"
+				}),
 				queueContainer(
 					withEnvVar("CONTAINER_CONCURRENCY", "0"),
 					withEnvVar("SERVING_READINESS_PROBE", `{"tcpSocket":{"port":8080,"host":"127.0.0.1"}}`),
@@ -567,15 +670,23 @@ func TestMakePodSpec(t *testing.T) {
 			}),
 	}, {
 		name: "with shell readiness probe",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withExecReadinessProbe([]string{"echo", "hello"}),
-			)
-		}),
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withExecReadinessProbe([]string{"echo", "hello"}),
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
-					withExecReadinessProbe([]string{"echo", "hello"})),
+				servingContainer(
+					func(container *corev1.Container) {
+						container.Image = "busybox@sha256:deadbeef"
+						container.ReadinessProbe = withExecReadinessProbe([]string{"echo", "hello"})
+					}),
 				queueContainer(
 					withEnvVar("CONTAINER_CONCURRENCY", "0"),
 					withEnvVar("SERVING_READINESS_PROBE", `{"tcpSocket":{"port":8080,"host":"127.0.0.1"}}`),
@@ -583,19 +694,29 @@ func TestMakePodSpec(t *testing.T) {
 			}),
 	}, {
 		name: "with http liveness probe",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withTCPReadinessProbe(),
-				withLivenessProbe(corev1.Handler{
-					HTTPGet: &corev1.HTTPGetAction{
-						Path: "/",
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+				LivenessProbe: &corev1.Probe{
+					Handler: corev1.Handler{
+						HTTPGet: &corev1.HTTPGetAction{
+							Path: "/",
+						},
 					},
-				}),
-			)
-		}),
+				},
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
+				servingContainer(
+					func(container *corev1.Container) {
+						container.Image = "busybox@sha256:deadbeef"
+					},
 					withLivenessProbe(corev1.Handler{
 						HTTPGet: &corev1.HTTPGetAction{
 							Path: "/",
@@ -613,17 +734,26 @@ func TestMakePodSpec(t *testing.T) {
 			}),
 	}, {
 		name: "with tcp liveness probe",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withTCPReadinessProbe(),
-				withLivenessProbe(corev1.Handler{
-					TCPSocket: &corev1.TCPSocketAction{},
-				}),
-			)
-		}),
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+				LivenessProbe: &corev1.Probe{
+					Handler: corev1.Handler{
+						TCPSocket: &corev1.TCPSocketAction{},
+					}}}},
+			),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
+				servingContainer(
+					func(container *corev1.Container) {
+						container.Image = "busybox@sha256:deadbeef"
+					},
 					withLivenessProbe(corev1.Handler{
 						TCPSocket: &corev1.TCPSocketAction{
 							Port: intstr.FromInt(v1.DefaultUserPort),
@@ -634,105 +764,98 @@ func TestMakePodSpec(t *testing.T) {
 					withEnvVar("CONTAINER_CONCURRENCY", "0"),
 				),
 			}),
-	}, {
-		name: "with /var/log collection",
-		rev: revision("bar", "foo", withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-				)
-			}),
-		oc: metrics.ObservabilityConfig{
-			EnableVarLogCollection: true,
-		},
-		want: podSpec(
-			[]corev1.Container{
-				userContainer(),
-				queueContainer(
-					withEnvVar("CONTAINER_CONCURRENCY", "1"),
-					withEnvVar("ENABLE_VAR_LOG_COLLECTION", "true"),
-					withInternalVolumeMount(),
-				),
-			},
-			func(podSpec *corev1.PodSpec) {
-				podSpec.Volumes = append(podSpec.Volumes, internalVolume)
-			},
-		),
-	}, {
-		name: "with graceful scaledown enabled",
-		rev: revision("bar", "foo", func(revision *v1.Revision) {
-			container(revision.Spec.GetContainer(),
-				withTCPReadinessProbe(),
-			)
-		}),
-		ac: autoscalerconfig.Config{
-			EnableGracefulScaledown: true,
-		},
-		want: podSpec(
-			[]corev1.Container{
-				userContainer(),
-				queueContainer(
-					withEnvVar("DOWNWARD_API_LABELS_PATH", fmt.Sprintf("%s/%s", podInfoVolumePath, metadataLabelsPath)),
-					withPodLabelsVolumeMount(),
-				),
-			},
-			func(podSpec *corev1.PodSpec) {
-				podSpec.Volumes = append(podSpec.Volumes, labelVolume)
-			},
-		),
 	}, {
 		name: "complex pod spec",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "busybox",
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
 			withContainerConcurrency(1),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
 			func(revision *v1.Revision) {
 				revision.ObjectMeta.Labels = map[string]string{
 					serving.ConfigurationLabelKey: "cfg",
 					serving.ServiceLabelKey:       "svc",
 				}
-				revision.Spec.GetContainer().Command = []string{"/bin/bash"}
-				revision.Spec.GetContainer().Args = []string{"-c", "echo Hello world"}
-				container(revision.Spec.GetContainer(),
-					withTCPReadinessProbe(),
-					withEnvVar("FOO", "bar"),
-					withEnvVar("BAZ", "blah"),
-				)
-				revision.Spec.GetContainer().Resources = corev1.ResourceRequirements{
-					Requests: corev1.ResourceList{
-						corev1.ResourceMemory: resource.MustParse("666Mi"),
-						corev1.ResourceCPU:    resource.MustParse("666m"),
-					},
-					Limits: corev1.ResourceList{
-						corev1.ResourceMemory: resource.MustParse("888Mi"),
-						corev1.ResourceCPU:    resource.MustParse("888m"),
-					},
-				}
-				revision.Spec.GetContainer().TerminationMessagePolicy = corev1.TerminationMessageReadFile
+				container(revision.Spec.GetContainer(), updateContainer())
 			},
 		),
 		want: podSpec(
 			[]corev1.Container{
-				userContainer(
+				servingContainer(
 					func(container *corev1.Container) {
-						container.Command = []string{"/bin/bash"}
-						container.Args = []string{"-c", "echo Hello world"}
-						container.Env = append([]corev1.EnvVar{{
-							Name:  "FOO",
-							Value: "bar",
-						}, {
-							Name:  "BAZ",
-							Value: "blah",
-						}}, container.Env...)
-						container.Resources = corev1.ResourceRequirements{
-							Requests: corev1.ResourceList{
-								corev1.ResourceMemory: resource.MustParse("666Mi"),
-								corev1.ResourceCPU:    resource.MustParse("666m"),
-							},
-							Limits: corev1.ResourceList{
-								corev1.ResourceMemory: resource.MustParse("888Mi"),
-								corev1.ResourceCPU:    resource.MustParse("888m"),
-							},
-						}
-						container.TerminationMessagePolicy = corev1.TerminationMessageReadFile
+						container.Image = "busybox@sha256:deadbeef"
+					},
+					getContainer(),
+					withEnvVar("K_CONFIGURATION", "cfg"),
+					withEnvVar("K_SERVICE", "svc"),
+				),
+				queueContainer(
+					withEnvVar("CONTAINER_CONCURRENCY", "1"),
+					withEnvVar("SERVING_SERVICE", "svc"),
+				)}),
+	}, {
+		name: "complex pod spec for multiple containers with container data to all containers",
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
+					ContainerPort: v1.DefaultUserPort,
+				}},
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}, {
+				Name:  sidecarContainerName,
+				Image: "ubuntu",
+			}, {
+				Name:  "sidecar-container-2",
+				Image: "alpine",
+			}}),
+			withContainerConcurrency(1),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}, {
+				ImageDigest: "ubuntu@sha256:deadbffe",
+			}, {
+				ImageDigest: "alpine@sha256:deadbfff",
+			}}),
+			func(revision *v1.Revision) {
+				revision.ObjectMeta.Labels = map[string]string{
+					serving.ConfigurationLabelKey: "cfg",
+					serving.ServiceLabelKey:       "svc",
+				}
+				for i := range revision.Spec.Containers {
+					container(&revision.Spec.Containers[i], updateContainer())
+				}
+			},
+		),
+		want: podSpec(
+			[]corev1.Container{
+				servingContainer(
+					func(container *corev1.Container) {
+						container.Image = "busybox@sha256:deadbeef"
+					},
+					getContainer(),
+					withEnvVar("K_CONFIGURATION", "cfg"),
+					withEnvVar("K_SERVICE", "svc"),
+				),
+				sidecarContainer(sidecarContainerName,
+					func(container *corev1.Container) {
+						container.Image = "ubuntu@sha256:deadbffe"
+					},
+					getContainer(),
+					withEnvVar("K_CONFIGURATION", "cfg"),
+					withEnvVar("K_SERVICE", "svc"),
+				),
+				sidecarContainer(sidecarContainerName2,
+					getContainer(),
+					func(container *corev1.Container) {
+						container.Name = "sidecar-container-2"
+						container.Image = "alpine@sha256:deadbfff"
 					},
 					withEnvVar("K_CONFIGURATION", "cfg"),
 					withEnvVar("K_SERVICE", "svc"),
@@ -742,6 +865,92 @@ func TestMakePodSpec(t *testing.T) {
 					withEnvVar("SERVING_SERVICE", "svc"),
 				),
 			}),
+	}, {
+		name: "complex pod spec for multiple containers with container data only to serving containers",
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
+					ContainerPort: 8888,
+				}},
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}, {
+				Name:  sidecarContainerName,
+				Image: "ubuntu",
+			}}),
+			withContainerConcurrency(1),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}, {
+				ImageDigest: "ubuntu@sha256:deadbffe",
+			}}),
+			func(revision *v1.Revision) {
+				revision.ObjectMeta.Labels = map[string]string{
+					serving.ConfigurationLabelKey: "cfg",
+					serving.ServiceLabelKey:       "svc",
+				}
+				container(revision.Spec.GetContainer(), updateContainer())
+			},
+		),
+		want: podSpec(
+			[]corev1.Container{
+				servingContainer(
+					func(container *corev1.Container) {
+						container.Image = "busybox@sha256:deadbeef"
+					},
+					getContainer(),
+					func(container *corev1.Container) {
+						container.Ports[0].ContainerPort = 8888
+					},
+					withEnvVar("PORT", "8888"),
+					withEnvVar("K_CONFIGURATION", "cfg"),
+					withEnvVar("K_SERVICE", "svc"),
+				),
+				sidecarContainer(sidecarContainerName,
+					func(container *corev1.Container) {
+						container.Image = "ubuntu@sha256:deadbffe"
+					},
+					withEnvVar("K_CONFIGURATION", "cfg"),
+					withEnvVar("K_SERVICE", "svc"),
+				),
+				queueContainer(
+					withEnvVar("CONTAINER_CONCURRENCY", "1"),
+					withEnvVar("SERVING_SERVICE", "svc"),
+					withEnvVar("USER_PORT", "8888"),
+					withEnvVar("SERVING_READINESS_PROBE", `{"tcpSocket":{"port":8888,"host":"127.0.0.1"}}`),
+				),
+			}),
+	}, {
+		name: "propertes allowed by the webhook are passed through",
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
+					ContainerPort: 8080,
+				}},
+				ReadinessProbe: withTCPReadinessProbe(v1.DefaultUserPort),
+			}}),
+			func(r *v1.Revision) {
+				// TODO - do this generically for all allowed properties
+				r.Spec.EnableServiceLinks = ptr.Bool(false)
+			}),
+		want: podSpec(
+			[]corev1.Container{
+				servingContainer(
+					withEnvVar("PORT", "8080"),
+					withEnvVar("K_REVISION", "bar"),
+				),
+				queueContainer(
+					withEnvVar("USER_PORT", "8080"),
+					withEnvVar("SERVING_READINESS_PROBE", `{"tcpSocket":{"port":8080,"host":"127.0.0.1"}}`),
+				),
+			},
+			func(p *corev1.PodSpec) {
+				p.EnableServiceLinks = ptr.Bool(false)
+			},
+		),
 	}}
 
 	for _, test := range tests {
@@ -749,9 +958,9 @@ func TestMakePodSpec(t *testing.T) {
 			quantityComparer := cmp.Comparer(func(x, y resource.Quantity) bool {
 				return x.Cmp(y) == 0
 			})
-			got, err := makePodSpec(test.rev, &logConfig, &traceConfig, &test.oc, &test.ac, &deploymentConfig)
+			got, err := makePodSpec(test.rev, &logConfig, &traceConfig, &test.oc, &deploymentConfig)
 			if err != nil {
-				t.Fatalf("makePodSpec returned error: %v", err)
+				t.Fatal("makePodSpec returned error:", err)
 			}
 			if diff := cmp.Diff(test.want, got, quantityComparer); diff != "" {
 				t.Errorf("makePodSpec (-want, +got) = %v", diff)
@@ -762,87 +971,155 @@ func TestMakePodSpec(t *testing.T) {
 
 func TestMissingProbeError(t *testing.T) {
 	if _, err := MakeDeployment(revision("bar", "foo"), &logConfig, &traceConfig,
-		&network.Config{}, &obsConfig, &asConfig, &deploymentConfig); err == nil {
+		&network.Config{}, &obsConfig, &deploymentConfig, &asConfig); err == nil {
 		t.Error("expected error from MakeDeployment")
 	}
 }
 
 func TestMakeDeployment(t *testing.T) {
 	tests := []struct {
-		name string
-		rev  *v1.Revision
-		want *appsv1.Deployment
+		name      string
+		rev       *v1.Revision
+		want      *appsv1.Deployment
+		dc        deployment.Config
+		acMutator func(*asconfig.Config)
 	}{{
 		name: "with concurrency=1",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:  servingContainerName,
+				Image: "busybox",
+				Ports: []corev1.ContainerPort{{
+					ContainerPort: 8888,
+				}},
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}, {
+				Name:  sidecarContainerName,
+				Image: "ubuntu",
+			}}),
 			withoutLabels,
 			withContainerConcurrency(1),
-			func(revision *v1.Revision) {
-				container(revision.Spec.GetContainer(),
-					withReadinessProbe(corev1.Handler{
-						TCPSocket: &corev1.TCPSocketAction{
-							Host: "127.0.0.1",
-							Port: intstr.FromInt(12345),
-						},
-					}),
-				)
-			},
-		),
-		want: makeDeployment(),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}, {
+				ImageDigest: "ubuntu@sha256:deadbffe",
+			}})),
+		want: appsv1deployment(),
 	}, {
 		name: "with owner",
 		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "ubuntu",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
 			withoutLabels,
 			withOwnerReference("parent-config"),
-			func(revision *v1.Revision) {
-				container(revision.Spec.GetContainer(),
-					withReadinessProbe(corev1.Handler{
-						TCPSocket: &corev1.TCPSocketAction{
-							Host: "127.0.0.1",
-							Port: intstr.FromInt(12345),
-						},
-					}),
-				)
-			},
-		),
-		want: makeDeployment(),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}})),
+		want: appsv1deployment(),
 	}, {
 		name: "with sidecar annotation override",
-		rev: revision("bar", "foo", withoutLabels, func(revision *v1.Revision) {
-			revision.ObjectMeta.Annotations = map[string]string{
-				sidecarIstioInjectAnnotation: "false",
-			}
-			container(revision.Spec.GetContainer(),
-				withReadinessProbe(corev1.Handler{
-					TCPSocket: &corev1.TCPSocketAction{
-						Host: "127.0.0.1",
-						Port: intstr.FromInt(12345),
-					},
-				}),
-			)
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "ubuntu",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}),
+			withoutLabels, func(revision *v1.Revision) {
+				revision.ObjectMeta.Annotations = map[string]string{
+					sidecarIstioInjectAnnotation: "false",
+				}
+			}),
+		want: appsv1deployment(func(deploy *appsv1.Deployment) {
+			deploy.Annotations[sidecarIstioInjectAnnotation] = "false"
+			deploy.Spec.Template.Annotations[sidecarIstioInjectAnnotation] = "false"
 		}),
-		want: makeDeployment(func(deploy *appsv1.Deployment) {
-			deploy.ObjectMeta.Annotations[sidecarIstioInjectAnnotation] = "false"
-			deploy.Spec.Template.ObjectMeta.Annotations[sidecarIstioInjectAnnotation] = "false"
+	}, {
+		name: "with ProgressDeadline override",
+		dc: deployment.Config{
+			ProgressDeadline: 42 * time.Second,
+		},
+		rev: revision("bar", "foo",
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "ubuntu",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
+			WithContainerStatuses([]v1.ContainerStatuses{{
+				ImageDigest: "busybox@sha256:deadbeef",
+			}}), withoutLabels),
+		want: appsv1deployment(func(deploy *appsv1.Deployment) {
+			deploy.Spec.ProgressDeadlineSeconds = ptr.Int32(42)
+		}),
+	}, {
+		name: "cluster initial scale",
+		acMutator: func(ac *asconfig.Config) {
+			ac.InitialScale = 10
+		},
+		rev: revision("bar", "foo",
+			withoutLabels,
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "ubuntu",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
+		),
+		want: appsv1deployment(func(deploy *appsv1.Deployment) {
+			deploy.Spec.Replicas = ptr.Int32(int32(10))
+		}),
+	}, {
+		name: "cluster initial scale override by revision initial scale",
+		acMutator: func(ac *asconfig.Config) {
+			ac.InitialScale = 10
+		},
+		rev: revision("bar", "foo",
+			withoutLabels,
+			withContainers([]corev1.Container{{
+				Name:           servingContainerName,
+				Image:          "ubuntu",
+				ReadinessProbe: withTCPReadinessProbe(12345),
+			}}),
+			func(revision *v1.Revision) {
+				revision.ObjectMeta.Annotations = map[string]string{autoscaling.InitialScaleAnnotationKey: "20"}
+			},
+		),
+		want: appsv1deployment(func(deploy *appsv1.Deployment) {
+			deploy.Spec.Replicas = ptr.Int32(int32(20))
+			deploy.Spec.Template.ObjectMeta.Annotations = map[string]string{autoscaling.InitialScaleAnnotationKey: "20"}
+			deploy.ObjectMeta.Annotations = map[string]string{autoscaling.InitialScaleAnnotationKey: "20"}
 		}),
 	}}
 
 	for _, test := range tests {
 		t.Run(test.name, func(t *testing.T) {
+			ac := &asconfig.Config{
+				InitialScale:          1,
+				AllowZeroInitialScale: false,
+			}
+			if test.acMutator != nil {
+				test.acMutator(ac)
+			}
 			// Tested above so that we can rely on it here for brevity.
 			podSpec, err := makePodSpec(test.rev, &logConfig, &traceConfig,
-				&obsConfig, &asConfig, &deploymentConfig)
+				&obsConfig, &deploymentConfig)
 			if err != nil {
-				t.Fatalf("makePodSpec returned error: %v", err)
+				t.Fatal("makePodSpec returned error:", err)
+			}
+			if test.want != nil {
+				test.want.Spec.Template.Spec = *podSpec
 			}
-			test.want.Spec.Template.Spec = *podSpec
 			got, err := MakeDeployment(test.rev, &logConfig, &traceConfig,
-				&network.Config{}, &obsConfig, &asConfig, &deploymentConfig)
+				&network.Config{}, &obsConfig, &test.dc, ac)
 			if err != nil {
-				t.Fatalf("got unexpected error: %v", err)
+				t.Fatal("got unexpected error:", err)
 			}
-			if diff := cmp.Diff(test.want, got, cmpopts.IgnoreUnexported(resource.Quantity{})); diff != "" {
-				t.Errorf("MakeDeployment (-want, +got) = %v", diff)
+			if diff := cmp.Diff(test.want, got, cmp.AllowUnexported(resource.Quantity{})); diff != "" {
+				t.Error("MakeDeployment (-want, +got) =", diff)
 			}
 		})
 	}
